# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: executed_tests.proto
# plugin: python-betterproto
# This file has been @generated

from dataclasses import dataclass
from datetime import datetime
from typing import List

import betterproto


@dataclass(eq=False, repr=False)
class Digest(betterproto.Message):
    """Taken directly from the bazel reapi, for compat"""

    hash: str = betterproto.string_field(1)
    """
    The hash. In the case of SHA-256, it will always be a lowercase hex string
    exactly 64 characters long. in expectation, this should always be the
    sha256
    """

    size_bytes: int = betterproto.int64_field(2)
    """The size of the blob, in bytes."""


@dataclass(eq=False, repr=False)
class ArtifactPointer(betterproto.Message):
    path: str = betterproto.string_field(1, group="pointer")
    artifact_name: str = betterproto.string_field(3)


@dataclass(eq=False, repr=False)
class TestResult(betterproto.Message):
    """
    The serialized record of single test execution This is the api that we
    should build tools around -- if people provide tracked tests
    """

    test_name: str = betterproto.string_field(1)
    """name of the test"""

    outputs: "TestOutputs" = betterproto.message_field(2)


@dataclass(eq=False, repr=False)
class TestOutputs(betterproto.Message):
    artifacts: List["ArtifactPointer"] = betterproto.message_field(1)
    """Files that are expected from a test"""

    exit_code: int = betterproto.int32_field(2)
    """exit code of the test"""


@dataclass(eq=False, repr=False)
class Invocation(betterproto.Message):
    """
    Highest level invocation for a set of tests -- must contain one or more
    test results any time we run any test(s), an invocation object is created
    """

    invoke_id: str = betterproto.string_field(1)
    rundate: datetime = betterproto.message_field(2)
    user: str = betterproto.string_field(3)
    repo: str = betterproto.string_field(4)
    branch: str = betterproto.string_field(5)
    hostname: str = betterproto.string_field(6)
    smelt_root: str = betterproto.string_field(7)
    executed_tests: List["TestResult"] = betterproto.message_field(8)



import pathlib
from typing import List, Optional
from pysmelt.proto.executed_tests import Invocation, TestResult
from pysmelt.interfaces.paths import SmeltPath
from dataclasses import dataclass


def most_recent_invoke_path() -> SmeltPath:
    """
    By default, we push the most recent invocation to this path

    """
    return SmeltPath("smelt-out/invocation.bin")


def most_recent_junit_path() -> SmeltPath:
    """
    By default, we push the most recent invocation to this path

    """
    return SmeltPath("smelt-out/tests.xml")


def get_previous_invocation() -> Invocation:
    with open(most_recent_invoke_path().to_abs_path(), "rb") as f:
        invbytes = f.read()
    return Invocation.FromString(invbytes)


def most_recent_tests_run() -> List[TestResult]:
    return get_previous_invocation().executed_tests


def read_log(test_result: TestResult) -> str:
    log_name = "smelt_log"
    log = next(
        (
            artifact
            for artifact in test_result.outputs.artifacts
            if artifact.artifact_name == log_name
        )
    )
    with open(log.path, "r") as f:
        return f.read()


def read_log_from_result(test: TestResult) -> Optional[str]:
    log_artifact = next(
        (
            artifact
            for artifact in test.outputs.artifacts
            if artifact.artifact_name == "command.log"
        ),
        None,
    )
    if log_artifact:
        logpath = pathlib.Path(log_artifact.path)
        if logpath.exists():
            return logpath.read_text()


@dataclass(frozen=True)
class IQL:
    """
    This is the "Invocation Query Layer" -- a helper class to get data out from Invocations

    """

    inv: Invocation

    @classmethod
    def from_previous(cls):
        return cls(inv=get_previous_invocation())

    def get_test(self, test_name: str) -> Optional[TestResult]:
        return next(
            (test for test in self.inv.executed_tests if test.test_name == test_name),
            None,
        )

    def get_log_content(self, test_name: str) -> Optional[str]:
        test = self.get_test(test_name)
        if test:
            return read_log_from_result(test)

    def get_tests_from_testgroup(
        self, test_group_name: str
    ) -> Optional[List[TestResult]]:
        test_group = self.get_test(test_group_name)
        if test_group:
            log_content = read_log(test_group)
            test_names = log_content.split("\n")
            return [
                test
                for test in (self.get_test(tn) for tn in test_names)
                if test is not None
            ]
        else:
            return None



from typing_extensions import Annotated, List
from pathlib import Path
import typer
import yaml
from pysmelt.interfaces.paths import SmeltPath
from pysmelt.output import smelt_console
from pysmelt.interfaces import SmeltTargetType
from pysmelt.proto.smelt_client.commands import CfgDocker, ConfigureSmelt
from pysmelt.rc import SmeltRcHolder
from pysmelt.smelt_muncher import parse_smelt
from pysmelt.output_utils import pretty_print_tests
from pysmelt.pygraph import create_graph
from pysmelt.serde import SafeDataclassDumper
from typing import Optional, Dict
from typer import Exit
from pysmelt.templates.template_rule import create_rule_target_from_template
from pysmelt.pysmelt import create_worker_binary

app = typer.Typer()


TlPath = Annotated[
    Path,
    typer.Argument(
        ...,
        exists=True,
        file_okay=True,
        dir_okay=False,
        writable=False,
        readable=True,
        resolve_path=True,
    ),
]

CommandPath = Annotated[
    Path,
    typer.Argument(
        exists=False,
        file_okay=True,
        dir_okay=False,
        writable=False,
        readable=True,
        resolve_path=True,
    ),
]


NewRulePath = Annotated[
    Path,
    typer.Argument(
        ...,
        exists=False,
        file_okay=True,
        dir_okay=False,
        writable=True,
        readable=True,
        resolve_path=True,
    ),
]


def validate_type(value: str):
    if value not in SmeltTargetType._value2member_map_:
        raise Exit(
            f'Invalid value for "--tt". Possible values are {", ".join(SmeltTargetType._value2member_map_.keys())}'  # type: ignore
        )
    return value


@app.command(
    help="Lowers an smelt file to a command file",
)
def lower(
    smelt_file: TlPath,
    output: CommandPath = Path("command.yaml"),
):
    typer.echo(f"Validating: {smelt_file}")

    _, commands = parse_smelt(test_list=SmeltPath.from_str(str(smelt_file)))
    with open(output, "w") as f:
        yaml.dump(commands, f, Dumper=SafeDataclassDumper, sort_keys=False)


@app.command(
    help="Executes an smelt file",
)
def execute(
    smelt_file: TlPath,
    tt: str = typer.Option("test", help="SMELT target type", callback=validate_type),
    target_name: Optional[str] = typer.Option(
        None, help="Target name -- if not provided, runs all the tests"
    ),
    test_only: bool = typer.Option(
        False,
        help="If set, assumes non-test commands have passed successfully and will not run them",
        is_flag=True,
    ),
    jobs: Optional[int] = typer.Option(
        None, "--jobs", help="max number of jobslots allowed"
    ),
    prepare_workspace: bool = typer.Option(
        False,
        help="If set, we will prepare the workspace for a distributed run -- Only build commands will run, and test commands will only create shell scripts.",
        is_flag=True,
    ),
):
    if jobs:
        SmeltRcHolder.set_jobs(jobs)

    def configure_cb(cfg: ConfigureSmelt) -> ConfigureSmelt:
        cfg.test_only = test_only
        cfg.prepare_workspace = prepare_workspace
        return cfg

    graph = create_graph(str(smelt_file), cfg_init=configure_cb)
    if target_name:
        graph.run_one_test_interactive(target_name)
    else:
        graph.run_all_typed_commands(tt)
    if prepare_workspace:
        create_worker_binary()


@app.command(
    help="Executes an smelt file using the docker executor",
)
def execute_docker(
    smelt_file: TlPath,
    img: str = typer.Option(
        ..., help="Docker image that each command will run inside of"
    ),
    tt: Optional[str] = typer.Option(
        "test", help="SMELT target type", callback=validate_type
    ),
    target_name: Optional[str] = typer.Option(
        None, help="Target name -- if not provided, runs all the tests"
    ),
    test_only: bool = typer.Option(
        False,
        help="If set, assumes non-test commands have passed successfully and will not run them",
        is_flag=True,
    ),
    jobs: Optional[int] = typer.Option(
        None, "--jobs", help="max number of jobslots allowed"
    ),
):

    if jobs:
        SmeltRcHolder.set_jobs(jobs)

    def configure_cb(cfg: ConfigureSmelt) -> ConfigureSmelt:
        cfg.test_only = test_only
        cfg.docker = CfgDocker()
        cfg.docker.image_name = img
        cfg.docker.additional_mounts = {}
        return cfg

    graph = create_graph(str(smelt_file), cfg_init=configure_cb)
    if target_name:
        graph.run_one_test_interactive(target_name)
    elif tt:
        graph.run_all_typed_commands(tt)
    else:
        graph.run_all_commands()


@app.command(
    help="validates a file is well formed",
)
def validate(
    smelt_file: TlPath,
):

    graph = create_graph(str(smelt_file))
    smelt_console.print(f"[green] {smelt_file.name} is valid")
    pretty_print_tests(graph)


@app.command(help="Create a new target def file at the provided path")
def init_rule(output: CommandPath):
    create_rule_target_from_template(str(output))
    smelt_console.log(f"Created template at path {output}")


def main():
    app()


if __name__ == "__main__":
    app()



from dataclasses import dataclass, field
from datetime import datetime
from typing import Dict, List, Optional, Tuple, cast
from pysmelt.output import smelt_console
import betterproto
from pysmelt.interfaces.analysis import (
    most_recent_invoke_path,
    most_recent_junit_path,
    read_log_from_result,
)
from pysmelt.interfaces.target import SmeltTargetType
from pysmelt.proto.smelt_telemetry import (
    CommandEvent,
    CommandFinished,
    CommandStarted,
    Event,
    ExecutionStart,
    InvokeEvent,
)
from pysmelt.proto.executed_tests import Invocation, TestResult


from junitparser.junitparser import (
    TestCase,
    TestSuite,
    JUnitXml,
    Skipped,
    Error,
    Failure,
)


@dataclass
class InvocationBuilder:
    """
    Subscriber to build an Invocation and JUnitXml object, that tracks outputs and other artifacts of tests


    """

    invoke_id: Optional[str] = None
    start: Optional[datetime] = None
    rundate: Optional[datetime] = None
    user: Optional[str] = None
    repo: Optional[str] = None
    branch: Optional[str] = None
    hostname: Optional[str] = None
    smelt_root: Optional[str] = None
    tests: List[Tuple[TestResult, str, datetime]] = field(default_factory=list)
    test_start: Dict[str, datetime] = field(default_factory=dict)

    def process_message(self, message: Event):
        (variant, event_payload) = betterproto.which_one_of(message, "et")
        if variant == "command":
            event_payload = cast(CommandEvent, event_payload)
            command_name = event_payload.command_ref
            (command_variant, command_payload) = betterproto.which_one_of(
                event_payload, "CommandVariant"
            )

            if command_variant == "started":
                command_payload = cast(CommandStarted, command_payload)
                self.test_start[command_name] = message.time

            if command_variant == "finished":
                command_payload = cast(CommandFinished, command_payload)

                self.tests.append(
                    (
                        TestResult(
                            test_name=command_name, outputs=command_payload.outputs
                        ),
                        command_payload.command_type,
                        message.time,
                    )
                )

            else:
                pass
        elif variant == "invoke":
            event_payload = cast(InvokeEvent, event_payload)
            (invoke_variant, invoke_payload) = betterproto.which_one_of(
                event_payload, "InvokeVariant"
            )

            if invoke_variant == "start":
                self.start = message.time
                invoke_payload = cast(ExecutionStart, invoke_payload)
                self.branch = invoke_payload.git_branch
                self.repo = invoke_payload.git_repo
                self.hostname = invoke_payload.hostname
                self.smelt_root = invoke_payload.smelt_root
                self.invoke_id = message.trace_id
                self.user = invoke_payload.username
            if invoke_variant == "done":
                self.rundate = message.time

        else:
            pass

    def create_invocation_object(self) -> Invocation:
        assert self.invoke_id, "invoke_id is required"
        assert self.rundate, "rundate is required"
        # these 4 are optional
        # assert self.user, "user is required"
        # assert self.repo, "repo is required"
        # assert self.branch, "branch is required"
        # assert self.hostname, "hostname is required"
        assert self.smelt_root, "smelt_root is required"

        results = [val[0] for val in self.tests]
        return Invocation(
            invoke_id=self.invoke_id,
            rundate=self.rundate,
            user=self.user,
            repo=self.repo,
            branch=self.branch,
            hostname=self.hostname,
            smelt_root=self.smelt_root,
            executed_tests=results,
        )

    def create_junit(self) -> JUnitXml:
        assert self.invoke_id, "invoke_id is required"
        assert self.rundate, "rundate is required"
        assert self.start, "start is required"
        assert self.smelt_root, "smelt_root is required"

        suite = TestSuite(name=self.invoke_id)
        if self.hostname:
            suite.hostname = self.hostname
        suite_duration = (self.rundate - self.start).total_seconds()
        suite.time = suite_duration

        for test in self.tests:
            tr, ttype, endtime = test
            duration = (endtime - self.test_start[tr.test_name]).total_seconds()
            case = TestCase(tr.test_name, ttype, duration)  # params are optional
            log = read_log_from_result(tr)
            if log:
                case.system_out = log
            if tr.outputs.exit_code != 0:
                case.result = [Failure(f"failed with exit code {tr.outputs.exit_code}")]
            suite.add_testcase(case)
        rv = JUnitXml()
        rv.add_testsuite(suite)
        return rv

    def write_invocation_and_junit(self):
        try:
            inv_obj = self.create_invocation_object().SerializeToString()
            with open(most_recent_invoke_path().to_abs_path(), "wb") as f:
                f.write(inv_obj)
            xml = self.create_junit()
            xml.write(most_recent_junit_path().to_abs_path(), pretty=True)
        except Exception as e:
            smelt_console.print("[yellow] Failed to create junit xml")



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: data.proto
# plugin: python-betterproto
# This file has been @generated

from dataclasses import dataclass
from datetime import datetime
from typing import (
    TYPE_CHECKING,
    Dict,
    Optional,
)

import betterproto
import betterproto.lib.google.protobuf as betterproto_lib_google_protobuf
import grpclib
from betterproto.grpc.grpclib_server import ServiceBase

from .. import executed_tests as _executed_tests__


if TYPE_CHECKING:
    import grpclib.server
    from betterproto.grpc.grpclib_client import MetadataLike
    from grpclib.metadata import Deadline


class SmeltErrorType(betterproto.Enum):
    CLIENT_ERROR = 0
    """Client caused error"""

    INTERNAL_ERROR = 1
    """Internal error -- anything that is thrown by the smelt runtime"""

    INTERNAL_WARN = 2
    """
    Internal warning -- anything that the smelt runtime wants to broadcast back
    """


@dataclass(eq=False, repr=False)
class Event(betterproto.Message):
    """Event flows from server -> client only"""

    time: datetime = betterproto.message_field(1)
    trace_id: str = betterproto.string_field(2)
    """A globally-unique ID (UUIDv4) of this trace. Required."""

    command: "CommandEvent" = betterproto.message_field(15, group="et")
    invoke: "InvokeEvent" = betterproto.message_field(16, group="et")
    error: "SmeltError" = betterproto.message_field(17, group="et")


@dataclass(eq=False, repr=False)
class CommandEvent(betterproto.Message):
    """CommandEvents covers activity happening on a per target basis"""

    command_ref: str = betterproto.string_field(1)
    """
    test def id this ref should be consistent for the same test being executed
    """

    scheduled: "CommandScheduled" = betterproto.message_field(4, group="CommandVariant")
    started: "CommandStarted" = betterproto.message_field(5, group="CommandVariant")
    cancelled: "CommandCancelled" = betterproto.message_field(6, group="CommandVariant")
    finished: "CommandFinished" = betterproto.message_field(7, group="CommandVariant")
    stdout: "CommandStdout" = betterproto.message_field(8, group="CommandVariant")
    profile: "CommandProfile" = betterproto.message_field(9, group="CommandVariant")
    skipped: "CommandSkipped" = betterproto.message_field(10, group="CommandVariant")


@dataclass(eq=False, repr=False)
class CommandScheduled(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class CommandStarted(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class CommandCancelled(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class CommandSkipped(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class CommandStdout(betterproto.Message):
    output: str = betterproto.string_field(1)


@dataclass(eq=False, repr=False)
class CommandFinished(betterproto.Message):
    outputs: "_executed_tests__.TestOutputs" = betterproto.message_field(1)
    command_type: str = betterproto.string_field(2)


@dataclass(eq=False, repr=False)
class CommandProfile(betterproto.Message):
    memory_used: int = betterproto.uint64_field(1)
    """memory used by the command, in bytes"""

    cpu_load: float = betterproto.float_field(2)
    """cpu load as a fraction -- 1.0 is one cpu, 2.0 is 2 cpus, etc"""

    time_since_start_ms: int = betterproto.uint64_field(3)
    """Number of milliseconds since profiling began"""


@dataclass(eq=False, repr=False)
class InvokeEvent(betterproto.Message):
    """InvokeEvent demarcates the start of a graph execution."""

    start: "ExecutionStart" = betterproto.message_field(5, group="InvokeVariant")
    done: "AllCommandsDone" = betterproto.message_field(6, group="InvokeVariant")
    set: "SetGraph" = betterproto.message_field(7, group="InvokeVariant")


@dataclass(eq=False, repr=False)
class ExecutionStart(betterproto.Message):
    smelt_root: str = betterproto.string_field(1)
    username: str = betterproto.string_field(2)
    hostname: str = betterproto.string_field(3)
    git_hash: str = betterproto.string_field(4)
    git_repo: str = betterproto.string_field(5)
    git_branch: str = betterproto.string_field(6)


@dataclass(eq=False, repr=False)
class AllCommandsDone(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class SetGraph(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class SmeltError(betterproto.Message):
    sig: "SmeltErrorType" = betterproto.enum_field(1)
    error_payload: str = betterproto.string_field(2)


class EventListenerStub(betterproto.ServiceStub):
    async def send_event(
        self,
        event: "Event",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None
    ) -> "betterproto_lib_google_protobuf.Empty":
        return await self._unary_unary(
            "/smelt_telemetry.EventListener/SendEvent",
            event,
            betterproto_lib_google_protobuf.Empty,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )

    async def send_outputs(
        self,
        executed_tests_test_result: "_executed_tests__.TestResult",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None
    ) -> "betterproto_lib_google_protobuf.Empty":
        return await self._unary_unary(
            "/smelt_telemetry.EventListener/SendOutputs",
            executed_tests_test_result,
            betterproto_lib_google_protobuf.Empty,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )


class EventListenerBase(ServiceBase):

    async def send_event(
        self, event: "Event"
    ) -> "betterproto_lib_google_protobuf.Empty":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def send_outputs(
        self, executed_tests_test_result: "_executed_tests__.TestResult"
    ) -> "betterproto_lib_google_protobuf.Empty":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def __rpc_send_event(
        self,
        stream: "grpclib.server.Stream[Event, betterproto_lib_google_protobuf.Empty]",
    ) -> None:
        request = await stream.recv_message()
        response = await self.send_event(request)
        await stream.send_message(response)

    async def __rpc_send_outputs(
        self,
        stream: "grpclib.server.Stream[_executed_tests__.TestResult, betterproto_lib_google_protobuf.Empty]",
    ) -> None:
        request = await stream.recv_message()
        response = await self.send_outputs(request)
        await stream.send_message(response)

    def __mapping__(self) -> Dict[str, grpclib.const.Handler]:
        return {
            "/smelt_telemetry.EventListener/SendEvent": grpclib.const.Handler(
                self.__rpc_send_event,
                grpclib.const.Cardinality.UNARY_UNARY,
                Event,
                betterproto_lib_google_protobuf.Empty,
            ),
            "/smelt_telemetry.EventListener/SendOutputs": grpclib.const.Handler(
                self.__rpc_send_outputs,
                grpclib.const.Cardinality.UNARY_UNARY,
                _executed_tests__.TestResult,
                betterproto_lib_google_protobuf.Empty,
            ),
        }



import yaml
import dataclasses


class SafeDataclassDumper(yaml.SafeDumper):
    def represent_data(self, data):
        if dataclasses.is_dataclass(data):
            return self.represent_dict(dataclasses.asdict(data))
        return super().represent_data(data)



from typing import List, Literal, Dict, Any, Optional, Tuple
from enum import Enum
from pysmelt.interfaces.runtime import RuntimeRequirements
from dataclasses import dataclass, asdict


from pysmelt.rc import SmeltRcHolder

CommandRef = str

# TODO: make this automatic from the targettype enum
CommandType = Literal["test", "stimulus", "build", "rebuild", "rerun"]


@dataclass
class Command:
    """
    The simplest unit of compute in smelt -- commands are the nodes that are scheduled and executed by the runtime

    Functionally, Command is a simple wrapper around a `bash` script
    """

    name: str
    target_type: CommandType
    script: List[str]
    """
    A list of bash commands that will be executed in sequence
    """
    dependencies: List[CommandRef]
    dependent_files: List[str]

    """
    Paths to the files that are expected to be created by this command 

    Anything here will be treated as an "artifact" 
    """
    outputs: List[str]
    runtime: RuntimeRequirements
    working_dir: str
    on_failure: Optional[CommandRef] = None

    @classmethod
    def from_dict(cls, data: Dict[str, Any]):
        name = data["name"]
        target_type = data["target_type"]
        script = data["script"]
        dependencies = data["dependencies"] if "dependencies" in data else []
        dependent_files = data["dependent_files"] if "dependent_files" in data else []
        outputs = data["outputs"] if "outputs" in data else []
        working_dir = (
            data["working_dir"]
            if "working_dir" in data
            else SmeltRcHolder.current_smelt_root()
        )

        runtime = RuntimeRequirements.from_dict(data["runtime"])

        return cls(
            name=name,
            target_type=target_type,
            script=script,
            dependent_files=dependent_files,
            dependencies=dependencies,
            outputs=outputs,
            runtime=runtime,
            working_dir=working_dir,
        )

    def to_dict(self) -> Dict[str, Any]:
        rv = asdict(self)

        return rv


class CStatus(Enum):
    PASS = "pass"
    FAIL = "failed"
    SKIPPED = "skipped"


CStatusStr = Literal[CStatus.PASS, CStatus.FAIL, CStatus.SKIPPED]  # ignore


@dataclass
class CResult:
    name: str
    status: CStatusStr



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: executed_tests.proto
# plugin: python-betterproto
from dataclasses import dataclass
from typing import Dict, List

import betterproto


@dataclass
class Digest(betterproto.Message):
    """Taken directly from the bazel reapi, for compat"""

    # The hash. In the case of SHA-256, it will always be a lowercase hex string
    # exactly 64 characters long. in expectation, this should always be the
    # sha256
    hash: str = betterproto.string_field(1)
    # The size of the blob, in bytes.
    size_bytes: int = betterproto.int64_field(2)


@dataclass
class ArtifactPointer(betterproto.Message):
    cas_hash: "Digest" = betterproto.message_field(1, group="pointer")
    path: str = betterproto.string_field(2, group="pointer")
    artifact_name: str = betterproto.string_field(3)


@dataclass
class TestMetaData(betterproto.Message):
    # User name of who owns the test
    test_owner: str = betterproto.string_field(1)
    # name of the design under test -- useful for tracking across different
    # testbenches
    design_under_test: str = betterproto.string_field(2)
    # Any extra data end-users might want to tag onto a test, for test
    # organization
    extras: Dict[str, str] = betterproto.map_field(
        3, betterproto.TYPE_STRING, betterproto.TYPE_STRING
    )


@dataclass
class TestResult(betterproto.Message):
    """
    The serialized record of single test execution A backend that stores this
    data should normalize the data in this object into a "TestDef" and a
    "TestInstantiation". I don't think it's complexity that should be exposed
    to the frontend
    """

    # same key from Invocation
    invoke_id: str = betterproto.string_field(1)
    # name of the test
    test_name: str = betterproto.string_field(2)
    # test definition id -- unique id to track the test across different
    # Invocations, useful for test history
    test_def_id: str = betterproto.string_field(3)
    # maps name of a file to a hash of the file contents To start off, we should
    # only support files here
    artifacts: List["ArtifactPointer"] = betterproto.message_field(4)
    # useful metadata for each test execution
    metadata: "TestMetaData" = betterproto.message_field(5)
    # cli reproduction -- we might want to track this a level above
    repro_command: str = betterproto.string_field(6)


@dataclass
class Invocation(betterproto.Message):
    # Highest level invocation for a set of tests -- must contain one or more
    # test results any time we run any test(s), an invocation object is created
    invoke_id: str = betterproto.string_field(1)
    rundate: str = betterproto.string_field(2)
    user: str = betterproto.string_field(3)
    repo: str = betterproto.string_field(4)
    branch: str = betterproto.string_field(5)
    hostname: str = betterproto.string_field(6)
    executed_tests: List["TestResult"] = betterproto.message_field(8)



from dataclasses import dataclass
from typing import ClassVar, Dict, List
from pysmelt.generators.procedural import dec_import_depth, inc_import_depth
from pysmelt.interfaces.command import Command
from pysmelt.interfaces.paths import SmeltPath, TempTarget
from pysmelt.interfaces.target import Target
from pysmelt.smelt_muncher import parse_smelt
from pysmelt.tracker import ImportTracker


def import_as_target(target_path: str) -> Target:
    inc_import_depth()
    tt = TempTarget.parse_string_smelt_target(
        target_path, ImportTracker.local_file_alias().path
    )
    chosen_name = tt.name
    if tt.file_path not in ImportTracker.imported_commands:
        targets, commands = parse_smelt(tt.file_path)
        # ImportTracker.imported_commands[tt.file_path] = commands
        if chosen_name not in targets:
            raise RuntimeError(f"Could not find command {target_path}")
    dec_import_depth()
    return ImportTracker.imported_targets[tt.file_path][chosen_name]



import time
from dataclasses import dataclass
from typing import Callable, Generator, List, Optional

import yaml
from pysmelt.pysmelt import PyController, PyEventStream

from pysmelt.interfaces import Command
from pysmelt.interfaces.paths import SmeltPath, SmeltPathFetcher
from pysmelt.proto.smelt_client.commands import (
    CfgDocker,
    CfgLocal,
    ConfigureSmelt,
    ProfilerCfg,
    ProfilingSelection,
)
from pysmelt.proto.smelt_telemetry import Event
from pysmelt.rc import SmeltRcHolder
from pysmelt.smelt_muncher import SmeltUniverse, create_universe
from pysmelt.subscribers import SmeltSub
from pysmelt.subscribers.error_handler import SmeltErrorHandler
from pysmelt.subscribers.invocation_builder import InvocationBuilder
from pysmelt.subscribers.output_collector import OutputConsole
from pysmelt.subscribers.retcode import RetcodeTracker
from pysmelt.subscribers.stdout import StdoutPrinter, StdoutSink


def default_cfg() -> ConfigureSmelt:
    rv = ConfigureSmelt()
    rc = SmeltRcHolder.current_rc()

    rv.job_slots = rc.jobs
    rv.smelt_root = rc.smelt_root
    rv.local = CfgLocal()
    rv.prof_cfg = ProfilerCfg()
    # sample memory
    rv.prof_cfg.prof_type = ProfilingSelection.SIMPLE_PROF
    # sample every 100 ms
    rv.prof_cfg.sampling_period = 100
    return rv


def maybe_get_message(
    listener: PyEventStream, blocking: bool = False
) -> Optional[Event]:
    try:
        if blocking:
            message = listener.pop_message_blocking()
            event = Event.FromString(message)
        else:
            message = listener.nonblocking_pop()
            if message is None:
                return None
            event = Event.FromString(message)

        return event
    except RuntimeError as e:
        """
        In expectation, we only return an error if the channel has been closed

        We don't have any paths above this function to handle that error, so we'll handle it here
        """

        print(e)
        return None


def spin_for_message(listener: PyEventStream, backoff: float = 0.2, time_out: int = 10):
    """
    Utility for spinning for a message
    """
    time_passed = 0
    while time_passed < time_out:
        message = maybe_get_message(listener, blocking=False)
        if message:
            return message
        else:
            time.sleep(backoff)
            time_passed += backoff

    raise RuntimeError(f"Before timeout of {time_out} expired!")


@dataclass
class PyGraph:
    universe: SmeltUniverse
    """ 
    holds all of the commands that are live in the graph 
    """
    controller: PyController
    """
    Handle to rust -- calls the function defined in the pyo3 bindings
    """

    retcode_tracker: RetcodeTracker

    additional_listeners: List[SmeltSub]

    def commands(self):
        return self.universe.top_level_commands

    def runloop(self, listener: PyEventStream):
        errhandler = SmeltErrorHandler()
        invbuilder = InvocationBuilder()
        with OutputConsole() as console:
            while not listener.is_done():
                # tbh, this could be async
                # but async interopt with rust is kind of experimental and i dont want to do it yet
                message = maybe_get_message(listener, blocking=False)
                if message:
                    self.retcode_tracker.process_message(message)
                    console.process_message(message)
                    errhandler.process_message(message)
                    invbuilder.process_message(message)
                    for other_listener in self.additional_listeners:
                        other_listener.process_message(message)
                if not message:
                    # add a little bit of backoff
                    time.sleep(0.01)
        invbuilder.write_invocation_and_junit()

    def console_runloop(
        self, test_name: str, listener: PyEventStream, sink: StdoutSink
    ) -> Generator[bool, None, None]:
        """
        Generator that will try to consume as many `Event` messages as possible

        All of the stdout from the "test_name" command will be given to the `sink`.
        By default, sink should just be print

        The yielded value will be
        """
        stdout_tracker = StdoutPrinter(test_name, sink)
        errhandler = SmeltErrorHandler()
        while True:
            # tbh, this could be async
            # but async interopt with rust is kind of experimental and i dont want to do it yet
            message = maybe_get_message(listener, blocking=False)
            if message:
                self.retcode_tracker.process_message(message)
                stdout_tracker.process_message(message)
                errhandler.process_message(message)
                for other_listener in self.additional_listeners:
                    other_listener.process_message(message)

            if not message:
                # add a little bit of backoff
                yield listener.is_done()

    def get_current_cfg(self) -> ConfigureSmelt:
        raw_cfg = self.controller.get_current_cfg()
        return ConfigureSmelt.FromString(raw_cfg)

    def reset(self):
        pass

    # self.retcode_tracker.reset()

    def run_one_test_interactive(self, name: str, sink: StdoutSink = print):
        """
        Runs a single test, with a "sink" handle to process all of the stdout for that specific command

        By default, this will just print stdout + stderr to the screen -- it looks like you're running the command interactively
        """
        self.reset()
        listener = self.controller.run_one_test(name)
        for is_done in self.console_runloop(name, listener, sink):
            if is_done:
                return
            time.sleep(0.1)

    def run_specific_commands(self, commands: List[Command]):
        self.reset()
        test_names = [command.name for command in commands]
        listener = self.controller.run_many_tests(test_names)
        self.runloop(listener)

    def run_all_typed_commands(self, maybe_type: str):
        self.reset()
        listener = self.controller.run_all_tests(maybe_type)
        self.runloop(listener)

    def run_all_commands(self):
        self.reset()
        toptests = self.universe.top_level_commands
        valid_commands = [
            command.name
            for command in toptests
            if command.target_type != "rebuild" and command.target_type != "rerun"
        ]
        listener = self.controller.run_many_tests(valid_commands)
        self.runloop(listener)

    def set_commands(self):
        """
        Initializes the list of commands that are visible to the smelt runtime


        If the list of commands are malformed -- e.g. syntax error in the yaml, then an error will be thrown
        """
        commands = self.universe.all_commands
        commands_as_str = yaml.safe_dump([command.to_dict() for command in commands])
        self.controller.set_graph(commands_as_str)

    @classmethod
    def init(cls, cfg: ConfigureSmelt, universe: SmeltUniverse):
        cfg_bytes = bytes(cfg)
        graph = PyController(cfg_bytes)
        rv = cls(
            universe=universe,
            controller=graph,
            retcode_tracker=RetcodeTracker(),
            additional_listeners=[],
        )
        rv.set_commands()
        return rv

    @classmethod
    def init_commands_only(cls, commands: List[Command]):
        cfg = default_cfg()
        top_path = SmeltPath.from_str(".")
        universe = SmeltUniverse(top_file=top_path, commands={top_path: commands})
        return cls.init(cfg, universe=universe)


def _create_cfg() -> ConfigureSmelt:
    cfg = default_cfg()
    return cfg


def create_graph(
    smelt_test_list: str,
    cfg_init: Optional[Callable[[ConfigureSmelt], ConfigureSmelt]] = None,
    default_rules_only: bool = False,
    file_fetcher: Optional[SmeltPathFetcher] = None,
) -> PyGraph:
    cfg = _create_cfg()
    if cfg_init:
        cfg = cfg_init(cfg)
    universe = create_universe(
        SmeltPath.from_str(smelt_test_list),
        default_rules_only=default_rules_only,
        file_fetcher=file_fetcher,
    )
    rv = PyGraph.init(cfg, universe)
    return rv


def create_graph_with_docker(smelt_test_list: str, cfg_docker: CfgDocker) -> PyGraph:
    def init_docker(cfg: ConfigureSmelt) -> ConfigureSmelt:
        cfg.docker = cfg_docker
        return cfg

    return create_graph(smelt_test_list, cfg_init=init_docker)



import importlib
import inspect
from typing import List, TypedDict, Dict, Type, Optional
from pysmelt.interfaces import Target
from pathlib import Path
from pysmelt.interfaces.command import Command
from pysmelt.interfaces.paths import (
    TempTarget,
    SmeltPathFetcher,
    local_smelt_path_fetcher,
)
from pysmelt.rc import SmeltRC
from pysmelt.path_utils import get_git_root
import sys
from importlib.util import spec_from_file_location, module_from_spec


class DocumentedTarget(TypedDict):
    target: Type[Target]
    doc: str


def get_all_files(targets_dir: Path) -> List[Path]:
    rules_files = []
    if targets_dir.exists() and targets_dir.is_dir():
        rule_files = [
            maybe_target_file for maybe_target_file in targets_dir.glob("**/*.py")
        ]
        return rule_files

    return rules_files


def get_all_targets(cfg: SmeltRC) -> Dict[str, DocumentedTarget]:
    rules_dir: Path = cfg.abs_rules_dir
    return _get_all_targets(rules_dir)


def get_default_targets(cfg: SmeltRC) -> Dict[str, DocumentedTarget]:
    return _get_all_targets(None)


def _get_all_targets(targets_dir: Optional[Path]) -> Dict[str, DocumentedTarget]:
    default_target_modules = ["pysmelt.default_targets"]
    classes = {}
    base_class_name = "Target"

    if targets_dir:
        all_paths = get_all_files(targets_dir=targets_dir)
        paths = default_target_modules + all_paths
    else:
        paths = default_target_modules

    for path in paths:
        try:

            if isinstance(path, str):
                module = importlib.import_module(str(path))
            elif isinstance(path, Path):
                module_name = path.stem  # get filename without extension
                spec = spec_from_file_location(module_name, str(path))
                module = module_from_spec(spec)
                spec.loader.exec_module(module)

            for name, cls in inspect.getmembers(module, inspect.isclass):
                if issubclass(cls, Target):
                    if name in classes:
                        raise ValueError(f"Duplicate target name: {name}")
                    if name != base_class_name:
                        classes[name] = cls
                        classes[name] = {"target": cls, "doc": inspect.getdoc(cls)}

        except ImportError as e:
            print(f"Failed to import rule definitions at {path} with error {e}")

    return classes


def import_procedural_testlist(py_path: str):
    spec = spec_from_file_location("__main__", py_path)

    module = module_from_spec(spec)

    spec.loader.exec_module(module)

    return module
